import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.HashMap;

/**
 * Ristinolla / Menu
 *
 * Valikot/menut, joilla ristinollaa (ruudukkoa) hallinnoidaan.
 *
 * @author Samuli Kaipiainen
 * @version 1.0
 */

public class Menu extends JMenuBar
{
	// komennettava ruudukko
	private Ruudukko ruudukko;

	// menut
	private JMenu peli = new JMenu("Peli");
	private JMenu peliPelaaja[] = { new JMenu("Pelaaja 1 / Rastit"), new JMenu("Pelaaja 2 / Pallot") };
	private JMenu peliPelaajaMenu[] = { new JMenu(), new JMenu() };
	private JMenu asetukset = new JMenu("Asetukset");
	private JMenu ohje = new JMenu("Help");

	// menuitemit

	private JMenuItem peliUusi = new JMenuItem("Uusi peli");
	private JMenuItem peliPois = new JMenuItem("Pois");

	private JMenuItem asetuksetNollaa = new JMenuItem("Nollaa ruudukon sijainti ja zoomaus");
	private JMenuItem asetuksetApuviivat = new JCheckBoxMenuItem("Apuviivat");
	private JMenuItem asetuksetJatkovoitto = new JCheckBoxMenuItem("Peli jatkuu voiton jlkeen");

	private JMenuItem ohjeApua = new JMenuItem("Apua");

	// MenuItem-PelaajaClass -kytkkset; Map koska tm ei ole tietokonepelaajan toteutus
	private HashMap menupelaaja = new HashMap();

	/**
	 * Luo ristinolla-menun
	 *
	 * @param r ruudukko, jota menu komentaa
	 */
	public Menu(Ruudukko r)
	{
		// ruudukko talteen
		this.ruudukko = r;

		// kivomman nkinen menu
		this.setBorder(BorderFactory.createEmptyBorder());
		// this.setBackground(Color.WHITE); // taustavri ei periydy menuihin eik menuitemeihin, ph

		// rakennetaan menut
		this.add(peli);
		peli.add(peliUusi);
		peli.addSeparator();
		peli.add(peliPelaaja[0]);
		peli.add(peliPelaajaMenu[0]);
		peli.addSeparator();
		peli.add(peliPelaaja[1]);
		peli.add(peliPelaajaMenu[1]);
		peli.addSeparator();
		peli.add(peliPois);
		this.add(asetukset);
		asetukset.add(asetuksetNollaa);
		asetukset.add(asetuksetApuviivat);
		asetukset.add(asetuksetJatkovoitto);
		this.add(ohje);
		ohje.add(ohjeApua);

		// piiloitetaan tyhjt pelaajamenut (ei tosin tarvitsisi, koska ne korvataan kohta)
		peliPelaajaMenu[0].setVisible(false);
		peliPelaajaMenu[1].setVisible(false);

		// alt-pikanapit
		peli.setMnemonic('p');
		peliUusi.setMnemonic('u');
		peliPelaaja[0].setMnemonic('r');
		peliPelaaja[1].setMnemonic('o');
		peliPois.setMnemonic('p');
		asetukset.setMnemonic('a');
		asetuksetNollaa.setMnemonic('n');
		asetuksetApuviivat.setMnemonic('a');
		asetuksetJatkovoitto.setMnemonic('j');
		ohje.setMnemonic('h');
		ohjeApua.setMnemonic('a');

		// muut pikanapit
		peliUusi.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F2, 0));
		asetuksetNollaa.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_BACK_SPACE, 0));

		// haetaan asetusten tilat ruudukolta
		asetuksetApuviivat.setSelected(ruudukko.onkoApuviivat());
		asetuksetJatkovoitto.setSelected(ruudukko.onkoJatkovoitto());

		// katsomma ketk osallistuvat turnajaisiin
		Class[] pelaaja = ruudukko.etsiPelaajat();
		ButtonGroup pelaajaryhm[] = { new ButtonGroup(), new ButtonGroup() };

		// listn menut erikseen rasteille ja palloille, mutta viitataan samoihin pelaajiin
		if (pelaaja != null) for (int p = 0; p < 2; p++) for (int n = 0; n < pelaaja.length; n++) {
			// listn pelaaja-MenuItem menuun, nappularyhmn ja mappiin
			JRadioButtonMenuItem peliPelaajaPelaaja = new JRadioButtonMenuItem(pelaaja[n].getName());
			peliPelaaja[p].add(peliPelaajaPelaaja);
			pelaajaryhm[p].add(peliPelaajaPelaaja);
			menupelaaja.put(peliPelaajaPelaaja, pelaaja[n]);

			// jos kyseess on ruudukon nykyinen pelaaja, asetetaan valituksi,
			// sek vaihdetaan kyseisen pelaajan menu edellisen tilalle
			if (pelaaja[n].getName().equals(ruudukko.annaPelaaja(p + 1).getName())) {
				peliPelaajaPelaaja.setSelected(true);
				vaihdaPelaajaMenu(p);
			}

			// alt-pikanappi ekaan kirjaimeen
			peliPelaajaPelaaja.setMnemonic(pelaaja[n].getName().charAt(0));

			// eri tapahtumaksittelij rasteille ja palloille, kunon niin hankalaa
			if (p == 0) peliPelaajaPelaaja.addActionListener(new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					ruudukko.asetaPelaaja(1, (Class) menupelaaja.get(e.getSource()));
					vaihdaPelaajaMenu(0);
				}
			});
			if (p == 1) peliPelaajaPelaaja.addActionListener(new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					ruudukko.asetaPelaaja(2, (Class) menupelaaja.get(e.getSource()));
					vaihdaPelaajaMenu(1);
				}
			});
		}

		// tapahtumaksittelijt lopuille menuille

		peliUusi.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				ruudukko.uusiPeli();
			}
		});

		peliPois.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				ruudukko.lopeta();
			}
		});

		asetuksetNollaa.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				ruudukko.nollaaRuudukonSijainti();
			}
		});

		asetuksetApuviivat.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				ruudukko.vaihdaApuviivat();
			}
		});

		asetuksetJatkovoitto.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				ruudukko.vaihdaJatkovoitto();
			}
		});

		ohjeApua.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				System.out.println("Hjelp");
			}
		});
	}

	/**
	 * Vaihtaa menuun kyseisen pelaajan menun; poistaa edellisen ja lis uuden
	 *
	 * @param p kenen menu vaihdetaan; 0==rasti, 1==pallo
	 */
	private void vaihdaPelaajaMenu(int p)
	{
		// poistetaan vanha menu
		peli.remove(peliPelaajaMenu[p]);

		// pyydetn ruudukolta uusi menu...
		peliPelaajaMenu[p] = ruudukko.annaPelaajaMenu(p + 1);

		// ...sek kyseisen menun omaavan pelaajan nimi
		String nimi = (p + 1) + " " + ruudukko.annaPelaaja(p + 1).getName() + " / asetukset";

		// tehdn oma (disabloitu) menu, jos ei pelaajalla olekaan menua
		if (peliPelaajaMenu[p] == null) (peliPelaajaMenu[p] = new JMenu(nimi)).setEnabled(false);
		else peliPelaajaMenu[p].setText(nimi);

		// alt-pikanappi pelaajanumeroon
		peliPelaajaMenu[p].setMnemonic('1' + p);

		// ei-lihavoitu fontti (jostain kieroutuneesta syyst)
		peliPelaajaMenu[p].setFont(peliPelaajaMenu[p].getFont().deriveFont(Font.PLAIN));

		// mihin kohtaan menu listn (ei voi korvata ktsysti tai kysy mik on vanhan sijainti)
		int pos = (p + 1) * 3;

		// listn uusi menu
		peli.insert(peliPelaajaMenu[p], pos);
	}
}
