import javax.swing.*;
import java.awt.event.*;

/**
 * Ristinolla / main-luokka
 *
 * @author Samuli Kaipiainen
 * @version 1.0
 */

public class Ristinolla extends JFrame
{
	// ruudukko ja menu
	private Ruudukko ruudukko;
	private JMenuBar menu;

	/**
	 * Luo uuden peli-ikkunan, lis ikkunaan ruudukon ja menun
	 *
	 * @param leveys ruudukon leveys ruutuina
	 * @param korkeus ruudukon korkeus ruutuina
	 * @param ruudunkoko ruudukon yhden ruudun koko pikselein
	 */
	public Ristinolla(int leveys, int korkeus, int ruudunkoko)
	{
		// luodaan ruudukko sek menu
		this.ruudukko = new Ruudukko(leveys, korkeus, ruudunkoko);
		this.menu = new Menu(this.ruudukko);

		// listn ikkunaan ruudukko ja menut
		this.getContentPane().add(this.ruudukko);
		this.setJMenuBar(this.menu);

		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.setTitle("Ristinolla");
		this.pack();
	}

	public static void main(String[] args)
	{
		// luetaan ruudukon koot komentorivilt, tai kytetn oletuksia
		int leveys = 18, korkeus = 12, ruudunkoko = 14;

		System.out.println("usage: 'java Ristinolla [leveys korkeus] [ruudunkoko]'");

		try {
			if (args.length == 1) ruudunkoko = Integer.parseInt(args[0]);
			else if (args.length == 2 || args.length == 3) {
				leveys = Integer.parseInt(args[0]);
				korkeus = Integer.parseInt(args[1]);
				if (args.length == 3) ruudunkoko = Integer.parseInt(args[2]);
			} else if (args.length > 3) throw new NumberFormatException();

		} catch (NumberFormatException e) {
			System.out.println("\n" +
					   "        leveys      ruudukon leveys ruutuina\n" +
					   "        korkeus     ruudukon korkeus ruutuina\n" +
					   "        ruudunkoko  ruudukon yhden ruudun koko pikselein\n" +
					   "\n" +
					   "oletus: " + leveys + "x" + korkeus + " @ " + ruudunkoko);
			System.exit(1);
		}

		// kerrotaan kytettv ruudukon koko
		System.out.println("using: " + leveys + "x" + korkeus + " @ " + ruudunkoko);

		// uusi ristinolla
		Ristinolla xo = new Ristinolla(leveys, korkeus, ruudunkoko);
		xo.show();

		// ruudukko kyntiin
		xo.ruudukko.kynnist();
	}
}
