// importit asetusmenua varten
import java.awt.event.*;
import javax.swing.*;

/**
 * Ristinolla / tietokonepelaaja
 *
 * Minmax-alphabeta-tietokonepelaaja.
 *
 * @author Samuli Kaipiainen
 * @version 1.0
 */

public class Tietokone implements Pelaaja
{
	// oma asetusvalikkomme
	private Menu menu;

	// ruudukkomme
	private Ruudukko ruudukko;

	// ruudukon koko
	private int leveys, korkeus;

	// edellinen saatu siirto
	private int xedellinen, yedellinen;

	// pelialue-ruudukko-taulukko; 0==tyhj, 1==rasti/min/tietokone, 2==pallo/vastustaja/ihminen
	private byte[][] r;

	// ruudukon alpha-kanava; 0==ruutu ei viel mukana haussa, >0==ruutu mukana haussa (ks. alphaSiirto)
	private int[][] ralpha;

	// pisteenlasku-slicet; 6 eri aloitusreunaa, 4 eri suoraa;
	// mritelmt ks. generate_slices() ja generate_slice()
	private int[][][] slice;

	// menusta muutettavat arvot...

	// ruudukon alpha-kanava -karsinta pll/pois
	// (kanavaa pivitetn vaikka karsinta olisi pois plt)
	private boolean ruudukkoalpha = true;

	// minmaxin alpha-beta -karsinta pll/pois
	private boolean alphabeta = true;

	// rekursion syvyys (kertoo suoraan rekursiossa kytettvn hakusyvyyden;
	// 0 ei hae ollenkaan, 1 hakee yhden oman siirron phn, ...)
	private int maxsyvyys = 3;

	// ...ovat skeiset kolme muuttujaa

	// rekursiossa sukeltava hetkittinen pelitilanteen hyvyysarvo
	private int pisteet;

	// minmaxin lytm paras hyvyysarvo sek sen toteuttava siirto; int[3] {x, y, arvo}
	private int[] paras;

	// pisteiden laskussa talteen laitettu voittotilanne haun katkaisua varten
	private boolean voitto;

	/**
	 * Kertoo tietokoneelle ruudukon tiedot
	 *
	 * @param ruudukko liitytty ruudukko
	 * @param leveys ruudukon leveys ruutuina
	 * @param korkeus ruudukon korkeus ruutuina
	 */
	public void asetaRuudukko(Ruudukko ruudukko, int leveys, int korkeus)
	{
		// ruudukko sek ruudukon koko talteen
		this.ruudukko = ruudukko;
		this.leveys = leveys;
		this.korkeus = korkeus;

		// oma pikku asetusmenumme
		this.menu = new Menu();

		// kerrotaan ruudukolle menumme olemassaolosta
		this.ruudukko.asetaPelaajaMenu(this, this.menu);
	}

	/**
	 * Aloittaa uuden pelin
	 */
	public void uusiPeli()
	{
		// nollataan edellinen siirto
		this.xedellinen = this.yedellinen = -2;

		// nollataan mys pisteet
		this.pisteet = 0;

		// luodaan ruudukkotaulukko ja sen alpha-kanava
		this.r = new byte[leveys][korkeus];
		this.ralpha = new int[leveys][korkeus];

		// merkataan alpha-kanavaan suunnilleen keskelt yksi ruutu, jotta osataan mys aloittaa
		// peli (tai jos vastustaja on tyhm ja aloittaa kulmasta, voidaan jatkaa keskelle)
		int x = leveys / 2 + (int) (Math.random() * (leveys / 3)) - leveys / 6;
		int y = korkeus / 2 + (int) (Math.random() * (korkeus / 3)) - korkeus / 6;
		ralpha[x][y]++;

		// sek slicet (ja pivitetn yhteispisteet)
		this.slice = generate_slices();
	}

	/**
	 * Hypyytt tietokoneen vanhaan peliin mukaan kesken pelin
	 */
	public void jatkaPeli()
	{
		// aloitetaan uusi peli
		uusiPeli();

		// pyydetn ruudukolta ruudukon kopio
		byte[][] tempr = ruudukko.annaRuudukonKopio(this);

		// listn tehdyt siirrot tiedoksemme
		for (int x = 0; x < tempr.length; x++)
			for (int y = 0; y < tempr[x].length; y++)
				if (tempr[x][y] != 0) teeSiirto(x, y, tempr[x][y]);
	}

	/**
	 * Antaa Ruudukolle uuden siirron; laskee sen minmax-algoritmilla
	 *
	 * @param x edellisen siirron x-ruutu
	 * @param y edellisen siirron y-ruutu
	 * @return uusi siirto int[2] {x, y} -taulukkona
	 */
	public int[] siirto(int x, int y)
	{
		// jos ei olla saatu uusiPeli-kutsua, pit hypt mukaan vanhaan peliin
		if (r == null) jatkaPeli();

		// josko yritimme viimeksi virheellist siirtoa, jossain on hmminki
		if (x == xedellinen && y == yedellinen) return null;

		// ja edellinen siirto talteen virhetarkistusta varten
		xedellinen = x; yedellinen = y;

		// rekursoidaan paras siirto
		return etsiSiirto(x, y);
	}

	/**
	 * Etsii parhaan siirron nykyiselle pelitilanteelle (vastustajan siirron jlkeen)
	 *
	 * Kutsuu minmaxia, joka sukeltaa maxsyvyys-muuttujan kertomaan syvyyteen;
	 * 0 palauttaa pelkn nykyisen tilanteen hyvyysarvon, 1 hakee yhden oman siirron phn, ...
	 *
	 * @param x vastustajan siirron x-ruutu
	 * @param y vastustajan siirron y-ruutu
	 *
	 * @return lydetty siirto ja siirron hyvyysarvo; int[3] {x, y, hyvyysarvo}
	 */
	private int[] etsiSiirto(int x, int y)
	{
		// ticksit talteen
		long ticks = System.currentTimeMillis();

		// tehdn vastustajan skeinen siirto (jollei ole tyhj pyt)
		if (x != -1 && y != -1) teeSiirto(x, y, 2);

		// nollataan mahdollinen edellisen siirron voittotilanne
		// (vaikkei tarvitsisi, ja vli vain jos peli jatkuu viel voiton jlkeen)
		voitto = false;

		// nollataan paras lydetty siirto
		paras = new int[] {-1, -1, Integer.MIN_VALUE};

		// minmaxista saatava hyvyysarvo (joka on tosin tallessa mys paras[2]:ssa)
		int hyvyys;

		// lhdetn etsimn vastustajan siirrosta eteenpin (meidn vuoro)
		// kutsutaan joko alphabetallista tai alphabetatonta minmaxia
		if (alphabeta) hyvyys = max(0, Integer.MIN_VALUE, Integer.MAX_VALUE);
		else hyvyys = minmax(0, true);

		// tehdn lydetty paras siirto (jos ruudukossa oli tilaa)
		if (paras[0] != -1 && paras[1] != -1) teeSiirto(paras[0], paras[1], 1);

		// debugataan nykyinen pistetilanne ja minmaxilla lydetty paras odotettavissa oleva
		System.out.print("Tietokone: nyt " + pisteet + ", paras(" + maxsyvyys + ") " + hyvyys);

		// kerrotaan kauanko siirto kesti
		ticks = System.currentTimeMillis() - ticks;
		System.out.println("; " + ticks + " ms");

		// palautetaan lydetty paras siirto
		return paras;
	}

	/**
	 * MinMax-argoritmi, ky oleellisen pelipuun lpi asetettuun maxsyvyyteen asti,
	 * ilman alpha-beta -karsintaa; karsinnan kanssa erillisiss max- ja min-metodeissa.
	 *
	 * Pivittelee r[][]-ruudukkoa ja slicej lennossa, mutta palauttaa ne alkutilaan valmistuttuaan.
	 *
	 * Tallettaa kunkinhetkisen pistetilanteen pisteet-muuttujaan, ja ensimmisell hakusyvyydell
	 * lytyvt siihenastiset parhaat siirrot paras[]-taulukkoon.
	 *
	 * @param syvyys hakusyvyys; aluksi 0, maxsyvyys palauttaa suoraan hyvyysarvon
	 * @param min onko nyt minun (tietokone) vuoro; true kyll, false ei vaan vastustajan
	 * @return paras (max) tai huonoin (min) lydetty hyvyysarvo
	 */
	private int minmax(int syvyys, boolean min)
	{
		// jos ollaan pelipuun kyseisen haaran lehdess, palautetaan lehden hyvyysarvo
		// (katkaistaan haku mys jos joku voitti edellisell siirrolla)
		if (syvyys >= maxsyvyys || voitto) return hyvyysarvo();

		// tmn oksan paras min- tai max-arvo
		int minmax = Integer.MIN_VALUE;

		// samanarvoisten siirtojen satunnaistamislaskuri (random-threshold iknkuin)
		// (1, jotta en yksi tyhj ruutu -tilanteessa MIN_VALUE:n kanssa samanarvoinen
		// ainoa mahdollinen siirto valitaan; muutoin (h > minmax) nollaa 2:ksi)
		int thres = 1;

		// kydn lpi kaikki ruudut...
		for (int x = 0; x < leveys; x++)
			for (int y = 0; y < korkeus; y++) {
				// kokeillaan plotata haussa mukana oleviin tyhjiin ruutuihin
				if ((!ruudukkoalpha || ralpha[x][y] != 0) && r[x][y] == 0) {
					// tehdn testattava siirto
					teeSiirto(x, y, min ? 1 : 2);
					// ok, sukelletaan... seuraavalla vuorolla eri pelaajan vuoro kuin nyt
					int h = minmax(syvyys + 1, !min);
					// (knnetn vastustajan arvo, jotta vertailu toimii)
					if (!min) h = -h;
					// jos lydetn minulle suurempi tai vastustajalle pienempi HA, talteen
					if (h > minmax) {
						minmax = h;
						// paras siirto ja sen hyvyysarvo talteen
						if (syvyys == 0) {
							paras[2] = minmax;
							paras[0] = x;
							paras[1] = y;
						}
						// nollataan satunnaistamislaskuri uudelle h-arvolle
						thres = 2;
					// ...tai yht suuri, jolloin lototaan uusi siirto talteen
					} else if (h == minmax && syvyys == 0 && Math.random() < (1.0 / thres++)) {
						paras[0] = x;
						paras[1] = y;
					}
					// perutaan tehty siirto
					peruSiirto(x, y);
				}
			}

		// knnetn takaisin vastustajan knteinen min-arvo
		if (!min) minmax = -minmax;

		// ...ja palautetaan saatu hyvyysarvo alaspin
		return minmax;
	}

	/**
	 * MinMax-algoritmin max-osa, alpha-beta -karsinnalla
	 *
	 * Erilliset max- ja min-metodit ovat selkemmt sek pikkaisen nopsemmat kuin
	 * purkkayhdistetty minmax-metodi, joka tosin on lyhyempi ja yllpidettvmpi
	 *
	 * @param syvyys tmnhetkinen rekursiosyvyys, aluksi 0
	 * @param alpha pienin alpha
	 * @param beta suurin beta
	 * @return lydetty paras hyvyysarvo
	 */
	private int max(int syvyys, int alpha, int beta)
	{
		// jos ollaan pelipuun kyseisen haaran lehdess, palautetaan lehden hyvyysarvo
		// (katkaistaan haku mys jos joku voitti edellisell siirrolla)
		if (syvyys >= maxsyvyys || voitto) return hyvyysarvo();

		// tmn oksan paras max-arvo
		int max = Integer.MIN_VALUE;

		// alpha-beta -karsinta pll
		boolean karsinta = false;

		// samanarvoisten siirtojen satunnaistamislaskuri (random-threshold iknkuin)
		// (1, jotta en yksi tyhj ruutu -tilanteessa MIN_VALUE:n kanssa samanarvoinen
		// ainoa mahdollinen siirto valitaan; muutoin (h > minmax) nollaa 2:ksi)
		int thres = 1;

		// kydn lpi kaikki ruudut...
siirrot:	for (int x = 0; x < leveys; x++)
			for (int y = 0; y < korkeus; y++) {
				// kokeillaan plotata haussa mukana oleviin tyhjiin ruutuihin
				if ((!ruudukkoalpha || ralpha[x][y] != 0) && r[x][y] == 0) {
					// tehdn testattava siirto, maxissa minun vuoro
					teeSiirto(x, y, 1);
					// ok, sukelletaan...
					int h = min(syvyys + 1, alpha, beta);
					// jos lydetn minulle suurempi HA, talteen
					if (h > max) {
						max = h;
						// kasvaako alpha-raja
						if (max > alpha) alpha = max;
						// mentiink beta-rajan ulkopuolelle (tai rajalle)
						if (max >= beta) karsinta = true;
						// paras siirto ja sen hyvyysarvo talteen
						if (syvyys == 0) {
							paras[2] = max;
							paras[0] = x;
							paras[1] = y;
						}
						// nollataan satunnaistamislaskuri uudelle h-arvolle
						thres = 2;
					// ...tai yht suuri, jolloin lototaan uusi siirto talteen
					} else if (h == max && syvyys == 0 && Math.random() < (1.0 / thres++)) {
						paras[0] = x;
						paras[1] = y;
					}
					// perutaan tehty siirto
					peruSiirto(x, y);
					if (karsinta) break siirrot;
				}
			}

		// ...ja palautetaan saatu paras hyvyysarvo alaspin
		return max;
	}

	/**
	 * MinMax-algoritmin min-osa, vastaava kuin max-osa mutta iknkuin negatiivisesti
	 *
	 * (Paitsi, jos/kun etsitn omasta siirrosta eteenpin, min() kutsutaan vasta seuraavalla
	 * rekursiotasolla, jolloin parasta siirtoa ei tarvitse laittaa talteen eik siirtoja arpoa)
	 *
	 * @param syvyys tmnhetkinen rekursiosyvyys, aluksi 0
	 * @param alpha pienin alpha
	 * @param beta suurin beta
	 * @return lydetty huonoin hyvyysarvo
	 */
	private int min(int syvyys, int alpha, int beta)
	{
		if (syvyys >= maxsyvyys || voitto) return hyvyysarvo();

		// tmn oksan paras min-arvo
		int min = Integer.MAX_VALUE;

		boolean karsinta = false;

siirrot:	for (int x = 0; x < leveys; x++)
			for (int y = 0; y < korkeus; y++) {
				if ((!ruudukkoalpha || ralpha[x][y] != 0) && r[x][y] == 0) {
					// miniss vastustajan vuoro
					teeSiirto(x, y, 2);
					int h = max(syvyys + 1, alpha, beta);
					// jos lydetn vastustajalle pienempi HA, talteen
					if (h < min) {
						min = h;
						// pieneneek beta-raja
						if (min < beta) beta = min;
						// mentiink alpha-rajan ulkopuolelle (tai rajalle)
						// (1-syvyydell yhtsuuruus ei kelpaa, koska 0-syvyys
						// arpoisi tmn mahdollisesti huonomman siirron
						// -- if (syvyys == 1 ? min < alpha : min <= alpha) karsinta = true;
						// VAI, pitk miniss aina olla aidosti ulkopuolella?
						// Tietokone pelaa eri lailla, jos sallitaan yhtsuuruus,
						// joten oletetaan ett nin se pelaa paremmin
						if (min < alpha) karsinta = true;
					}
					peruSiirto(x, y);
					if (karsinta) break siirrot;
				}
			}

		// ...ja palautetaan saatu huonoin hyvyysarvo alaspin
		return min;
	}

	/**
	 * Tekee annetun siirron; pivitt ruudukon ja slicet
	 *
	 * @param x siirron x-ruutu
	 * @param y siirron y-ruutu
	 * @param merkki kenen siirto tehdn; 1==rasti 2==pallo
	 */
	private void teeSiirto(int x, int y, int merkki)
	{
		// nollataan voittotilanne (score_slice laittaa plle jos huomaa voittotilanteen)
		voitto = false;

		// pivtn ruudukko...
		r[x][y] = (byte) merkki;

		// ...alpha-kanava (hallinnoidaan vaikka ruudukon alpha-karsinta olisi pois plt,
		// jolloin karsinnan saa laitettua lennossa plle; vie tosin 15%..20% suoritusaikaa)...
		alphaSiirto(x, y, 1, true);

		// ...ja slicet (sek pisteet)
		update_slices(x, y, merkki);
	}

	/**
	 * Peruu tehdyn siirron; pivitt ruudukon ja slicet
	 *
	 * teeSiirto() voisi mys tallettaa siirron (koordinaatit+merkin) pinoon, jolloin peruSiirto()
	 * ei tarvitsisi parametreja, poppaisi vain tavarat pinosta; vaan toimii hyvin+nopeammin ninkin
	 * ...Kunhan vain perutaan aina viimeksi tehty siirto, muutoin pistelaskupinot menevt sekaisin
	 *
	 * @param x siirron x-ruutu
	 * @param y siirron y-ruutu
	 */
	private void peruSiirto(int x, int y)
	{
		// pivtn ruudukko, alpha-kanava (samalla steell kuin siirtoa tehdess) ja slicet
		r[x][y] = 0;
		alphaSiirto(x, y, 1, false);
		update_slices(x, y, 0);
	}

	/**
	 * Merkkaa ruudukon alpha-kanavaan annetun ruudun ja sen ymprivt, steen sisll olevat ruudut
	 * mukaan hakuun tai pois hausta (kasvattaa tai vhent niiden arvoa yhdell)
	 *
	 * @param x0 x-ruutu jonka ymprist expandoidaan
	 * @param y0 y-ruutu jonka ymprist expandataan
	 * @param sde ymprivien ruutujen lisyssde; 1 lis kaikki vierekkiset ruudut (9 ruutua)
	 * @param kasvata listnk==true vai vhennetnk==false alpha-arvoa
	 */
	private void alphaSiirto(int x0, int y0, int sde, boolean kasvata)
	{
		// merkataan ymprivt ruudut hakuun mukaan
		int x1 = x0 - sde;
		int x2 = x0 + sde;
		int y1 = y0 - sde;
		int y2 = y0 + sde;

		// pysytn ruudukon sisll kuitenkin
		if (x1 < 0) x1 = 0;
		if (x2 >= leveys) x2 = leveys - 1;
		if (y1 < 0) y1 = 0;
		if (y2 >= korkeus) y2 = korkeus - 1;

		// pivitetn ymprivt ruudut alpha-kanavaan
		for (int y = y1; y <= y2; y++)
			for (int x = x1; x <= x2; x++)
				if (kasvata) ralpha[x][y]++;
				else ralpha[x][y]--;
	}

	/**
	 * Laskee nykyisen pelitilanteen heuristisen hyvyysarvon (joka tosin onkin jo laskettu valmiiksi)
	 *
	 * @return tilanteen hyvyysarvo
	 */
	private int hyvyysarvo()
	{
		return pisteet;
	}

	/**
	 * Luo slicet (suoraviipaleet) pisteidenlaskua varten, lis mys slaissien pisteet yhteispisteisiin
	 * (kannattanee siis nollata yhteispisteet ennen slice-generointia, ettei tule offsetti)
	 *
	 * @return varsin kaikki slicet int[ryhm][numero][] -taulukkona:
	 *
	 * int[0] lnsi-it vasemmasta reunasta, alkaa (0,0) alaspin
	 * int[1] pohjoinen-etel ylreunasta, alkaa (0,0) oikealle
	 * int[2] luode-kaakko ylreunasta, alkaa (0,0) oikealle
	 * int[3] luode-kaakko vasemmasta reunasta, alkaa (0,1) alaspin
	 * int[4] lounas-koillinen alareunasta, alkaa (0,korkeus-1) oikealle
	 * int[5] lounas-koillinen vasemmasta reunasta, alkaa (0,1) alaspin
	 *
	 * ...Nihin slice-mappauksiin ei sitten kosketa, tai mikn ei en toimi;
	 * ks. update_slices(), et varmaan halua alkaa miettimn sit uudestaan?)
	 */
	private int[][][] generate_slices()
	{
		// kaikki slicet, 6 eri aloitusreunaa (4 eri suuntaa)
		int slice[][][] = new int[6][][];

		// lnsi-it vasemmasta reunasta
		slice[0] = new int[korkeus][];
		for (int y = 0; y < korkeus; y++) slice[0][y] = generate_slice(0, y, 1, 0);

		// pohjoinen-etel ylreunasta
		slice[1] = new int[leveys][];
		for (int x = 0; x < leveys; x++) slice[1][x] = generate_slice(x, 0, 0, 1);

		// luode-kaakko ylreunasta
		slice[2] = new int[leveys][];
		for (int x = 0; x < leveys; x++) slice[2][x] = generate_slice(x, 0, 1, 1);

		// luode-kaakko vasemmasta reunasta
		slice[3] = new int[korkeus - 1][];
		for (int y = 1; y < korkeus; y++) slice[3][y - 1] = generate_slice(0, y, 1, 1);

		// lounas-koillinen alareunasta
		slice[4] = new int[leveys][];
		for (int x = 0; x < leveys; x++) slice[4][x] = generate_slice(x, korkeus - 1, 1, -1);

		// lounas-koillinen vasemmasta reunasta
		slice[5] = new int[korkeus - 1][];
		for (int y = 0; y < korkeus - 1; y++) slice[5][y] = generate_slice(0, y, 1, -1);

		return slice;
	}

	/**
	 * Luo yhden slicen; lis slicen pisteet yhteispisteisiin
	 *
	 * Nyt on sitten niin jotta, joka sliceen varataan tilaa niin monille pisteille, kuin on tarvetta
	 * rekursioon sukeltamisen kannalta, siisp tarkalleen slicen ruutupituuden + 1 verran (enemp
	 * siirtoja yhteen sliceen ei voi tehd); aina kun slice pivitetn update_slice() merkki!=0,
	 * tuota pistepinoa kasvatetaan, ja siirtoa peruttaessa merkki==0 pinoa vhennetn;
	 * l shell tmn kanssa tai itkuvirsi on vlttmtn (ja kyll, pistepinon pitisi
	 * olla erillisess taulukossa/pinossa, vaan eip nyt vain ole)
	 *
	 * @param x0 alku-x
	 * @param y0 alku-y
	 * @param dx x-lisys; tulee olla 0, 1 tai -1
	 * @param dy y-lisys; tulee olla 0, 1 tai -1
	 * @return luotu slice int[] -taulukkona:
	 *         { merkit ..., terminaattori (-1), slicen pisteet ..., pistepino-osoitin }
	 *         ... miss pino-osoitin osoittaa aluksi -1 jlkeiseen ensimmiseen pistealkioon
	 */
	private int[] generate_slice(int x0, int y0, int dx, int dy)
	{
		// varataan reilusti tilaa
		int[] slice = new int[leveys > korkeus ? leveys : korkeus];
		int n = 0;

		// kopioidaan vain, ei pakata tyhji, niin pivitys on simppeli
		for (int x = x0, y = y0; x >= 0 && y >= 0 && x < leveys && y < korkeus; x += dx, y += dy)
			slice[n++] = r[x][y];

		// dumpataan ylimrinen tila ja varataan tilaa terminaattorille sek pisteille
		int[] temp_slice = slice;
		// n ruutua, -1, n + 1 pisteytyst, pistepino-osoitin
		slice = new int[n + 1 + n + 2];
		System.arraycopy(temp_slice, 0, slice, 0, n);

		// terminaattori process_slice varten
		slice[n] = -1;

		// slicen alkupisteet (tyhjlle ruudukolle lienee 0)
		slice[n + 1] = score_slice(slice);

		// listn slicen pisteet yhteispisteisiin
		pisteet += slice[n + 1];

		// pistepino-osoitin osoittamaan ensimmiseen pistealkioon (jossa on slicen alkupisteet)
		slice[slice.length - 1] = n + 1;

		return slice;
	}

	/**
	 * Pivitt annetun ruudun plt kulkeviin sliceihin uuden merkin sek pisteet ja yhteispisteet
	 *
	 * Miten tm olisi pitnyt tehd, jottei tarvitsisi sotkea noiden [3][y-x-1] -indeksien kanssa?
	 * Ruudukon jokaiselle ruudulle int[][] -taulukko, jossa on kyseisen ruudun plt kulkevat
	 * slicet, joita on joka ruudulle tasan 4 kappaletta. Helppoa ja tehokasta (paitsi muistia
	 * kuluu hieman enemmn), vaan enp ala ttkn en myhemmin viisaampana muuttelemaan.
	 *
	 * @param x x-ruutu
	 * @param y y-ruutu
	 * @param merkki uusi merkki (0==tyhj 1==rasti 2==pallo)
	 */
	private void update_slices(int x, int y, int merkki)
	{
		// lnsi-it
		update_slice(slice[0][y], x, merkki);

		// pohjoinen-etel
		update_slice(slice[1][x], y, merkki);

		// luode-kaakko
		if (x >= y) update_slice(slice[2][x-y], y, merkki);
		else update_slice(slice[3][y-x-1], x, merkki);

		// lounas-koillinen
		if (x >= korkeus-1-y) update_slice(slice[4][x-korkeus+1+y], korkeus-1-y, merkki);
		else update_slice(slice[5][y+x], x, merkki);
	}

	/**
	 * Pivitt annetun slicen yhden merkin sek slicen pisteet ja yhteispisteet
	 *
	 * @param slice pivitettv slice
	 * @param n monesko merkki
	 * @param merkki uusi merkki
	 */
	private void update_slice(int[] slice, int n, int merkki)
	{
		if (n >= slice.length / 2 - 1) System.out.println("update_slice: n yli slicen, ei huv");

		// slicen pistepinolaskurin indeksi
		int pp = slice.length - 1;

		// pivitetn uusi merkki sliceen
		slice[n] = merkki;

		// nin katsomme, olemmeko tekemss vai perumassa siirtoa
		if (merkki != 0) {
			// kasvatetaan pistepinolaskuria, pivitetn uudet pisteet
			pisteet -= slice[slice[pp]++];
			pisteet += slice[slice[pp]] = score_slice(slice);
		} else {
			// vhennetn pistepinolaskuria, palautetaan vanhat pisteet
			pisteet -= slice[slice[pp]--];
			pisteet += slice[slice[pp]];
		}
	}

	/**
	 * Pisteytt slicen; laskee hyvyysarvon sille
	 *
	 * Antaa paremmat pisteet vastustajan voittotilanteille, mist seuraa jotta koko
	 * minmax-algoritmi toimii parhaiten parittomilla hakusyvyyksill (meidn oma siirto
	 * viimeisen); parillisilla syvyyksill tietokone on turhan ylipuolustava
	 *
	 * @param slice (-1)-terminoitu slice-taulukko
	 * @return slicen pisteet; 0 tasan, + rastit voitolla, - pallot voitolla
	 */
	private int score_slice(int[] slice)
	{
		// slicen hyvyysarvo, edellinen merkki (1==x, 2==o, 3==alku)
		int h = 0, m = 3;
		// samoja merkkej perkkin, tyhj tilaa ennen ja jlkeen
		int p = 0, ta = 0, tb = 0;

		// kydn slice lpi (-1)-arvoon asti
		for (int n = 0; m != -1; n++) {

			// ensin etsitn kaikki edelliset merkit (rakentuva suora)
			for (; slice[n] == m; n++) p++;

			// sitten suoran jlkeen oleva tyhj tila
			for (; slice[n] == 0; n++) tb++;

			// lopuksi lasketaan tuon suoran pisteet
			// jos alun tyhj ptk tai ei ole tilaa suoralle, ei maksa vaivaa
			if (m != 3 && ta + p + tb >= 5) {

				// jos joku voitti jo, laitetaan talteen (ja katkaistaan haku)
				if (p >= 5) voitto = true;
				// kerrotaan kahdella, niin vliin mahtuu molemmilla puolilla tilaa -pisteet
				p <<= 1;
				// jos molemmilla puolilla tilaa, hyv juttu
				if (ta > 0 && tb > 0) p++;
/*
				// -- vanha holtiton pistelaskukaava
				// varma voittotilanne saa villiksi
				if (p >= 5) p += 2;
				// varma voitto vastustajalle, jos pariton hakusyvyys (muutoin ylipuolustus)
				if (p >= 4 && m == 2) p += 2;
				// sovelletaan nerokasta pisteenlaskukaavaa (ylivuoto jos p->~15)
				int points = 1 << (p - 1 << 1);
*/
				int points = 0;
				// valkataan pisteet suoran pituuden mukaan
				switch (p) {
					// ox____
					case 2: points = 1; break;
					// _x___
					case 3:
					// oxx___
					case 4: points = 2; break;
					// _xx__
					case 5:
					// oxxx__
					case 6: points = 8; break;
					// _xxx_ (-2 varma voitto, jos seuraavana saman pelaajan vuoro)
					case 7: if (m == 1) points = 32;
						else points = 256;
						break;
					// oxxxx_ (-1 varma voitto, jos seuraavana saman pelaajan vuoro)
					case 8: if (m == 1) points = 40; // 32+8, oli 512?!
						else points = 1024; // sama kuin _xxxx_ ...
						break;
					// _xxxx_ (-1 varma voitto)
					case 9: if (m == 1) points = 1024;
						else points = 1536;
						break;
					// oxxxxx (voitto)
					case 10: points = 4096; break;
					// _xxxxx_
					case 11:
					// oxxxxxx
					case 12:
					default: points = p << 9;
				}

				// listn tai vhennetn pisteet hyvyysarvosta
				if (m == 1) h += points;
				else h -= points;
			}

			// edellinen merkki talteen
			m = slice[n];
			// nollataan materiaali
			p = 1; ta = tb; tb = 0;
		}

		return h;
	}

	/**
	 * Rekursoitujen siirtojen int-talletuspino, kasvattaa kokoaan jos tarvis
	 *
	 * ...Nonniin, tt ei sitten tarvitakaan ollenkaan, kun tein pistepinot erikseen
	 * joka sliceen; jtnp tnne kiusallani kuitenkin nyttmn ptevlt
	 * (lisksi tmn pikku pinon tekeminen oli suunnilleen kivointa koko tyss)
	 */
	private class Pino
	{
		private int[] stuff;
		private int at = 0;

		private Pino(int size)
		{
			stuff = new int[size];
		}

		private void put(int i)
		{
			// kaksinkertaistetaan koko jos ei mahdu
			if (at == stuff.length) {
				int[] temp = stuff;
				stuff = new int[stuff.length * 2];
				for (int n = 0; n < temp.length; n++) stuff[n] = temp[n];
			}

			stuff[at++] = i;
		}

		private int get()
		{
			return stuff[--at];
		}

		private boolean stuffed()
		{
			return at != 0;
		}
	}

	/**
	 * Tietokoneen asetusvalikko, josta voi vaihtaa hakusyvyyden sek laittaa
	 * minmaxin alpha-beta -karsinnan tai ruudukon alpha-kanava -karsinnan plle/pois
	 */
	private class Menu extends JMenu
	{
		// menuitemit
		private JMenuItem mruudukkoalpha = new JCheckBoxMenuItem("Ruudukon alpha-kanava -karsinta");
		private JMenuItem malphabeta = new JCheckBoxMenuItem("MinMaxin alpha-beta -karsinta");
		private JMenuItem[] mhakusyvyys;
		private ButtonGroup mhakusyvyydet;

		/**
		 * Luo asetusmenun (ei tarvitse nime; pmenu asettaa sen)
		 */
		public Menu()
		{
			// tapahtumaksittelij kaikille menuitemeille
			MenuKuuntelija menukuuntelija = new MenuKuuntelija();

			// asetetaan asetusten tilat tietokoneen nykyisiin (oletus)tiloihin
			this.mruudukkoalpha.setSelected(ruudukkoalpha);
			this.malphabeta.setSelected(alphabeta);

			// alt-pikanapit
			mruudukkoalpha.setMnemonic('r');
			malphabeta.setMnemonic('m');

			// listn tapahtumksittelijt
			mruudukkoalpha.addActionListener(menukuuntelija);
			malphabeta.addActionListener(menukuuntelija);

			// rakennetaan menut
			this.add(mruudukkoalpha);
			this.add(malphabeta);

			// reilusti hakusyvyyksi (vika jrkev on 5)
			this.mhakusyvyys = new JRadioButtonMenuItem[8];
			this.mhakusyvyydet = new ButtonGroup();

			// hakusyvyysmenuitemit
			for (int n = 0; n < mhakusyvyys.length; n++) {
				mhakusyvyys[n] = new JRadioButtonMenuItem("Hakusyvyys " + (n + 1));
				if (n < 3) mhakusyvyys[n].setText(mhakusyvyys[n].getText() + " (nopea)");
				if (n > 4) mhakusyvyys[n].setText(mhakusyvyys[n].getText() + " (hiiidas)");
				if (n + 1 == maxsyvyys) mhakusyvyys[n].setSelected(true);
				mhakusyvyys[n].setMnemonic('1' + n);
				mhakusyvyys[n].addActionListener(menukuuntelija);
				mhakusyvyydet.add(mhakusyvyys[n]);
				this.add(mhakusyvyys[n]);
			}
		}

		/**
		 * Tapahtumaksittelij kaikille menuitemeille
		 */
		private class MenuKuuntelija implements ActionListener
		{
			public void actionPerformed(ActionEvent e)
			{
				if (e.getSource() == mruudukkoalpha)
					ruudukkoalpha = !ruudukkoalpha;

				if (e.getSource() == malphabeta)
					alphabeta = !alphabeta;

				for (int n = 0; n < mhakusyvyys.length; n++)
					if (e.getSource() == mhakusyvyys[n])
						maxsyvyys = n + 1;
			}
		}
	}
}
